from collections.abc import MutableMapping
from contextlib import suppress
from operator import itemgetter
import sqlite3

# Code from the talk [Build powerful, new data structures with Python's abstract base classes]
# (https://www.youtube.com/watch?v=S_ipdVNSFlo) by [Raymond Hettinger](https://twitter.com/raymondh).


class SQLDict(MutableMapping):
    def __init__(self, dbname, items=[], **kwargs):
        self.dbname = dbname
        self.conn = sqlite3.connect(dbname)
        c = self.conn.cursor()
        with suppress(sqlite3.OperationalError):
            c.execute("CREATE TABLE Dict (key text, value text)")
            c.execute("CREATE UNIQUE INDEX KIndx ON Dict (key)")
        self.update(items, **kwargs)

    def __setitem__(self, key, value):
        if key in self:
            del self[key]
        with self.conn as c:
            c.execute("INSERT INTO  Dict VALUES (?, ?)", (key, value))

    def __getitem__(self, key):
        c = self.conn.execute("SELECT value FROM Dict WHERE Key=?", (key,))
        row = c.fetchone()
        if row is None:
            raise KeyError(key)
        return row[0]

    def __delitem__(self, key):
        if key not in self:
            raise KeyError(key)
        with self.conn as c:
            c.execute("DELETE FROM Dict WHERE key=?", (key,))

    def __len__(self):
        return next(self.conn.execute("SELECT COUNT(*) FROM Dict"))[0]

    def __iter__(self):
        c = self.conn.execute("SELECT key FROM Dict")
        return map(itemgetter(0), c.fetchall())

    def __repr__(self):
        return (
            f"{type(self).__name__}(dbname={self.dbname!r}, items={list(self.items())})"
        )

    def vacuum(self):
        with self.conn as c:
            c.execute("VACUUM;")

    def close(self):
        self.conn.close()



import base64
from collections.abc import MutableMapping
from contextlib import suppress
import os
from typing import Callable

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC  # , ByteString


class FileDict(MutableMapping):

    """A password can be provided, default is from env.

    The salt is expected as `str` to make it easy to set as an environment variable.
    After being processed it returns `bytes`

    The encoder/decoder can be overridden. The default value expected is a string so the default
    encoder/decoder is string.encode() and string.decode().

    Example:
    _______

    >>> from cryptofiledict import FileDict
    >>> enc_filedict = FileDict("cryptest")

    >>> import pickle
    >>> import math

    >>> enc_filedict.encoder = lambda x: pickle.dumps(x)
    >>> enc_filedict.decoder = lambda x: pickle.loads(x)

    >>> enc_filedict["myfunc"] = math.cos
    >>> enc_filedict["myfunc"]
    <function math.cos(x, /)>

    >>> enc_filedict["myfunc"](23)
    -0.5328330203333975

    >>> import json
    >>> enc_filedict.encoder = lambda x: json.dumps(x).encode()
    >>> enc_filedict.decoder = lambda x: json.loads(x.decode())

    >>> enc_filedict["mydict"] = {"a": 12}
    >>> enc_filedict["mydict"]
    {'a': 12}

    >>> enc_filedict["mydict"]["a"]
    12

    _______

    """

    def __init__(
        self,
        dirname: str,
        pairs=(),
        password: str = None,
        salt: str = None,
        encoder: Callable = lambda x: x.encode(),
        decoder: Callable = lambda x: x.decode(),
        **kwargs,
    ):
        self.dirname: str = dirname
        # get PASS from env if it is not provided
        self.password: bytes = os.getenv(
            "PASS"
        ).encode() if not password else password.encode()
        self.salt: bytes = base64.decodebytes(
            os.getenv("SALT").encode().decode("unicode_escape").encode()
        ) if not salt else salt.encode()  # provided as base64 string (not bytes)
        self.kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=self.salt,
            iterations=100000,
            backend=default_backend(),
        )
        self.key = base64.urlsafe_b64encode(self.kdf.derive(self.password))
        self.fernet = Fernet(self.key)

        # define encoder / decoder
        self.encoder = encoder
        self.decoder = decoder

        with suppress(FileExistsError):
            os.mkdir(dirname)
        self.update(pairs, **kwargs)

    def __setitem__(self, key, value):
        fullname = os.path.join(self.dirname, key)
        value = self.encoder(value)
        with open(fullname, "wb") as f:
            value = self.fernet.encrypt(value)
            f.write(value)

    def __getitem__(self, key):
        fullname = os.path.join(self.dirname, key)
        try:
            with open(fullname, "rb") as f:
                value = f.read()
                decrypt = self.fernet.decrypt(value)
                return self.decoder(decrypt)
                # return f.read()
        except FileNotFoundError:
            raise KeyError(key) from None

    def __delitem__(self, key):
        fullname = os.path.join(self.dirname, key)
        try:
            os.remove(fullname)
        except FileNotFoundError:
            raise KeyError(key) from None

    def __len__(self):
        return len(os.listdir(self.dirname))

    def __iter__(self):
        return iter(os.listdir(self.dirname))

    def __repr__(self):
        return f"FileDict{tuple(self.items())}"



import base64
import os
import secrets
import tarfile
from collections.abc import MutableMapping
from contextlib import suppress
from glob import glob
from typing import Callable

# import cryptography
from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from methodtools import lru_cache


class FileDict(MutableMapping):

    """A password can be provided, default is from env.

    The salt is generated for each entry. The salt is saved in the filename
    separated by a quadruple underscore (so don't use that in your variable names,
    you can use double and triple underscores).

    After being processed it returns a `ByteString`

    The encoder/decoder can be overridden. The default value expected is a string so the default
    encoder/decoder is string.encode() and string.decode().

    Example:
    _______

    >>> from flightcryptofiledict import FileDict
    >>> enc_filedict = FileDict("cryptest")

    >>> import pickle
    >>> import math

    >>> enc_filedict.encoder = lambda x: pickle.dumps(x)
    >>> enc_filedict.decoder = lambda x: pickle.loads(x)

    >>> enc_filedict["myfunc"] = math.cos
    >>> enc_filedict["myfunc"]
    <function math.cos(x, /)>

    >>> enc_filedict["myfunc"](23)
    -0.5328330203333975

    >>> import json
    >>> enc_filedict.encoder = lambda x: json.dumps(x).encode()
    >>> enc_filedict.decoder = lambda x: json.loads(x.decode())

    >>> enc_filedict["mydict"] = {"a": 12}
    >>> enc_filedict["mydict"]
    {'a': 12}

    >>> enc_filedict["mydict"]["a"]
    12
    
    _______

    """

    def __init__(
        self,
        dirname: str,
        pairs=(),
        password: str = None,
        encoder: Callable = lambda x: x.encode(),
        decoder: Callable = lambda x: x.decode(),
        **kwargs,
    ):
        self.dirname: str = dirname
        # get PASS from env if it is not provided
        self.password: bytes = os.getenv(
            "PASS"
        ).encode() if not password else password.encode()
        # self.kdfgen = lambda newsalt: PBKDF2HMAC(
        #     algorithm=hashes.SHA256(),
        #     length=32,
        #     salt=newsalt,
        #     iterations=100000,
        #     backend=default_backend(),
        # )
        # self.keygen = lambda kdf: base64.urlsafe_b64encode(kdf.derive(self.password))
        # # self.key = base64.urlsafe_b64encode(self.kdf.derive(self.password))
        # self.fernetgen = lambda key: Fernet(key)

        # define encoder / decoder
        self.encoder = encoder
        self.decoder = decoder

        with suppress(FileExistsError):
            os.mkdir(dirname)
        self.update(pairs, **kwargs)

    # 234375 == using 15mb of memory to cache fernet objects
    @lru_cache(maxsize=234375)
    def fernetgen(self, newsalt):
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=newsalt,
            iterations=100000,
            backend=default_backend(),
        )

        key = base64.urlsafe_b64encode(kdf.derive(self.password))

        return Fernet(key)

    def __setitem__(self, key, value):
        # delete file if exists. Since the salt is generated
        # on creation, the same key is saved with different filenames
        # so just writing on it will not overwrite
        fullpath = os.path.join(self.dirname, key)
        fullnames: list = glob(f"{fullpath}____*")
        if len(fullnames) > 0:
            # for f in fullnames:
            os.remove(fullnames[0])

        # key adaptation
        salt = secrets.token_urlsafe(64)
        # print("salt:", salt)
        key = f"{key}____{salt}"
        fullname = os.path.join(self.dirname, key)
        value = self.encoder(value)

        # new hasher generation
        newsalt = salt.encode()
        fernet = self.fernetgen(newsalt)
        #  kdf = self.kdfgen(newsalt)
        #  # use passkey, variable key is used before as the dictionary key
        #  passkey = self.keygen(kdf)
        #  fernet = self.fernetgen(passkey)
        # salt = fullname.split("____")[1]

        with open(fullname, "wb") as f:
            value = fernet.encrypt(value)
            f.write(value)

    def __getitem__(self, key):
        # key adaptation
        fullpath = os.path.join(self.dirname, key)
        try:
            fullname = glob(f"{fullpath}____*")[0]
        except IndexError:
            raise KeyError("key not found")

        # new hashser generation
        salt = fullname.split("____")[1]
        # print("salt:", salt)
        newsalt = salt.encode()
        fernet = self.fernetgen(newsalt)
        # kdf = self.kdfgen(newsalt)
        # key = self.keygen(kdf)
        # fernet = self.fernetgen(key)

        try:
            with open(fullname, "rb") as f:
                value = f.read()
                decrypt = fernet.decrypt(value)
                return self.decoder(decrypt)
                # return f.read()
        except FileNotFoundError:
            raise KeyError(key) from None

    def __delitem__(self, key):
        # key adaptation
        fullpath = os.path.join(self.dirname, key)
        fullname = glob(f"{fullpath}____*")[0]

        # fullname = os.path.join(self.dirname, key)
        try:
            os.remove(fullname)
        except FileNotFoundError:
            raise KeyError(key) from None

    def __len__(self):
        return len(os.listdir(self.dirname))

    def __iter__(self):
        files = [file.split("____")[0] for file in os.listdir(self.dirname)]
        return iter(files)

    def __repr__(self):
        # files = [file.split("____")[0] for file in os.listdir(self.dirname)]
        return f"FileDict{tuple(self.items())}"

    def compress(self, filename: str = None, **kwargs):
        """Create tar.gz file with the same name as the dictionary."""

        if not filename:
            filename = self.dirname

        with tarfile.open(filename + ".tar.gz", mode="w:gz", **kwargs) as tar:
            tar.add(self.dirname, arcname=filename, recursive=True)



from collections.abc import MutableMapping
from contextlib import suppress
import os


class FileDict(MutableMapping):
    def __init__(self, dirname, pairs=(), **kwargs):
        self.dirname = dirname
        with suppress(FileExistsError):
            os.mkdir(dirname)
        self.update(pairs, **kwargs)

    def __setitem__(self, key, value):
        fullname = os.path.join(self.dirname, key)
        with open(fullname, "w") as f:
            f.write(value)

    def __getitem__(self, key):
        fullname = os.path.join(self.dirname, key)
        try:
            with open(fullname) as f:
                return f.read()
        except FileNotFoundError:
            raise KeyError(key) from None

    def __delitem__(self, key):
        fullname = os.path.join(self.dirname, key)
        try:
            os.remove(fullname)
        except FileNotFoundError:
            raise KeyError(key) from None

    def __len__(self):
        return len(os.listdir(self.dirname))

    def __iter__(self):
        return iter(os.listdir(self.dirname))

    def __repr__(self):
        return f"FileDict{tuple(self.items())}"



import base64
import os
import secrets
import sqlite3
from collections.abc import MutableMapping
from contextlib import suppress
from operator import itemgetter
from typing import ByteString, Callable, Optional, Union

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from methodtools import lru_cache


class SQLDict(MutableMapping):

    # 234375 == using 15mb of memory to cache fernet objects
    # cache_n = 234375

    def __init__(
        self,
        dbname,
        items=[],
        password: Optional[str] = None,
        encoder: Callable = lambda x: x.encode(),
        decoder: Callable = lambda x: x.decode(),
        cache_size: Union[float, int] = None,
        **kwargs,
    ):
        self.dbname = dbname
        self.conn = sqlite3.connect(dbname)
        c = self.conn.cursor()
        self.password: ByteString = os.getenv(
            "PASS"
        ).encode() if not password else password.encode()
        self.encoder = encoder
        self.decoder = decoder

        # 6.4e-05 is de size in MB of a fernet object that can
        # encrypt / decrypt data (measured with sys.getsizeof)
        # 234375 == using 15mb of memory to cache fernet objects
        # if cache_size:
        #     nonlocal cache_n
        #     cache_n = cache_size / 6.4e-05

        with suppress(sqlite3.OperationalError):
            c.execute("CREATE TABLE Dict (key text, value blob, salt text)")
            c.execute("CREATE UNIQUE INDEX KIndx ON Dict (key)")

        self.update(items, **kwargs)

    # 234375 == using 15mb of memory to cache fernet objects
    @lru_cache(maxsize=234375)
    def _fernetgen(self, newsalt):
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=newsalt,
            iterations=100000,
            backend=default_backend(),
        )

        key = base64.urlsafe_b64encode(kdf.derive(self.password))

        return Fernet(key)

    def __setitem__(self, key, value):
        if key in self:
            del self[key]

        salt = secrets.token_urlsafe(64)
        value = self.encoder(value)
        newsalt = salt.encode()
        fernet = self._fernetgen(newsalt)
        value = fernet.encrypt(value)
        with self.conn as c:
            c.execute("INSERT INTO  Dict VALUES (?, ?, ?)", (key, value, salt))

    def __getitem__(self, key):
        c = self.conn.execute("SELECT value, salt FROM Dict WHERE Key=?", (key,))
        row = c.fetchone()
        if row is None:
            raise KeyError(key)
        value = row[0]
        salt = row[1]
        newsalt = salt.encode()
        fernet = self._fernetgen(newsalt)
        value = fernet.decrypt(value)
        return self.decoder(value)

    def __delitem__(self, key):
        if key not in self:
            raise KeyError(key)
        with self.conn as c:
            c.execute("DELETE FROM Dict WHERE key=?", (key,))

    def __len__(self):
        return next(self.conn.execute("SELECT COUNT(*) FROM Dict"))[0]

    def __iter__(self):
        c = self.conn.execute("SELECT key FROM Dict")
        return map(itemgetter(0), c.fetchall())

    def __repr__(self):
        return f"{type(self).__name__}(dbname={self.dbname!r})"  # , items={list(self.items())})"

    def vacuum(self):
        with self.conn as c:
            c.execute("VACUUM;")

    def close(self):
        self.conn.close()
